package org.topbraid.spin.arq;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.topbraid.spin.model.Ask;
import org.topbraid.spin.model.Construct;
import org.topbraid.spin.model.Describe;
import org.topbraid.spin.model.ElementList;
import org.topbraid.spin.model.FunctionCall;
import org.topbraid.spin.model.NamedGraph;
import org.topbraid.spin.model.Optional;
import org.topbraid.spin.model.SPINFactory;
import org.topbraid.spin.model.Select;
import org.topbraid.spin.model.Union;
import org.topbraid.spin.model.Variable;
import org.topbraid.spin.system.ExtraPrefixes;
import org.topbraid.spin.system.SPINPreferences;
import org.topbraid.spin.vocabulary.SP;
import org.topbraid.spin.vocabulary.SPIN;
import org.topbraid.spin.vocabulary.SPL;

import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.graph.Triple;
import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.query.SortCondition;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.RDFList;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.sparql.core.TriplePath;
import com.hp.hpl.jena.sparql.core.Var;
import com.hp.hpl.jena.sparql.core.VarExprList;
import com.hp.hpl.jena.sparql.expr.Expr;
import com.hp.hpl.jena.sparql.expr.ExprFunction;
import com.hp.hpl.jena.sparql.expr.ExprVar;
import com.hp.hpl.jena.sparql.expr.NodeValue;
import com.hp.hpl.jena.sparql.path.P_Alt;
import com.hp.hpl.jena.sparql.path.P_Link;
import com.hp.hpl.jena.sparql.path.P_Mod;
import com.hp.hpl.jena.sparql.path.P_Reverse;
import com.hp.hpl.jena.sparql.path.P_Seq;
import com.hp.hpl.jena.sparql.path.Path;
import com.hp.hpl.jena.sparql.syntax.Element;
import com.hp.hpl.jena.sparql.syntax.ElementAssign;
import com.hp.hpl.jena.sparql.syntax.ElementFilter;
import com.hp.hpl.jena.sparql.syntax.ElementGroup;
import com.hp.hpl.jena.sparql.syntax.ElementNamedGraph;
import com.hp.hpl.jena.sparql.syntax.ElementOptional;
import com.hp.hpl.jena.sparql.syntax.ElementPathBlock;
import com.hp.hpl.jena.sparql.syntax.ElementTriplesBlock;
import com.hp.hpl.jena.sparql.syntax.ElementUnion;
import com.hp.hpl.jena.sparql.syntax.Template;
import com.hp.hpl.jena.sparql.syntax.TemplateTriple;
import com.hp.hpl.jena.vocabulary.RDF;
import com.hp.hpl.jena.vocabulary.XSD;


/**
 * Takes a ARQ SPARQL Query as input and creates a corresponding SPIN
 * data structure from it.
 * 
 * @author Holger Knublauch
 */
public class ARQ2SPIN {
	
	private boolean addPrefixes;
	
	private Model model;
	
	private Model symbolsModel = SPL.getModel();
	
	private String varNamespace = SP.NS;
	
	private Map<String,Resource> var2Resource = new HashMap<String,Resource>();
	

	/**
	 * Constructs a new ARQ2SPIN engine for a given Model,
	 * equivalent with <code>ARQ2SPIN(model, true)</code>.
	 * @param model  the Model to operate on
	 */
	public ARQ2SPIN(Model model) {
		this(model, true);
	}	

	
	/**
	 * Constructs a new ARQ2SPIN engine for a given Model.
	 * @param model  the Model to operate on
	 * @param addPrefixes  true to also let the system add missing
	 *                     prefixes mentioned in SPARQL expressions
	 *                     (e.g. the afn namespace if afn:now() is used)
	 */
	public ARQ2SPIN(Model model, boolean addPrefixes) {
		this.model = model;
		this.addPrefixes = addPrefixes;
		
		// Pre-populate with named variables
		StmtIterator it = model.listStatements(null, SP.varName, (RDFNode)null);
		while(it.hasNext()) {
			Resource variable = it.nextStatement().getSubject();
			if(variable.isURIResource()) {
				if(SPINPreferences.get().isCreateURIVariables() ||
						variable.getURI().startsWith(SP.NS + "arg") ||
						SPIN.NS.equals(variable.getNameSpace())) {
					Variable var = (Variable) variable.as(Variable.class);
					String name = var.getName();
					var2Resource.put(name, var);
				}
			}
		}
	}


	@SuppressWarnings("unchecked")
	private void addDescribeProperties(Query arq, Resource spinQuery) {
		if(!arq.isQueryResultStar()) {
			List<Resource> members = new LinkedList<Resource>();
			Iterator vars = arq.getResultVars().iterator();
			while(vars.hasNext()) {
				String varName = (String) vars.next();
				Resource variable = getVariable(varName);
				members.add(variable);
			}
			Iterator uris = arq.getResultURIs().iterator();
			while(uris.hasNext()) {
				Node uriNode = (Node) uris.next();
				members.add(model.getResource(uriNode.getURI()));
			}
			spinQuery.addProperty(SP.resultNodes, model.createList(members.iterator()));
		}
	}
	
	
	@SuppressWarnings("unchecked")
	private void addGroupBy(Query arq, Resource spinQuery) {
		VarExprList namedExprs = arq.getGroupBy();
		Iterator vit = namedExprs.getVars().iterator();
		if(vit.hasNext()) {
			List<RDFNode> members = new LinkedList<RDFNode>();
			while(vit.hasNext()) {
		        Var var = (Var)vit.next();
		        Expr expr = namedExprs.getExpr(var) ;
		        if(expr == null) {
					String varName = (String) var.getName();
					Resource variable = getVariable(varName);
					members.add(variable);
		        }
		        else {
		        	throw new IllegalArgumentException("Expressions not supported in GROUP BY");
		        }
			}
			spinQuery.addProperty(SP.groupBy, model.createList(members.iterator()));
		}
	}


	@SuppressWarnings("unchecked")
	private void addNamedGraphClauses(Query arq, Resource spinQuery) {
		Iterator graphURIs = arq.getGraphURIs().iterator();
		while(graphURIs.hasNext()) {
			String graphURI = (String)graphURIs.next();
			spinQuery.addProperty(SP.from, model.getResource(graphURI));
		}
		
		Iterator namedGraphURIs = arq.getNamedGraphURIs().iterator();
		while(namedGraphURIs.hasNext()) {
			String namedGraphURI = (String)namedGraphURIs.next();
			spinQuery.addProperty(SP.fromNamed, model.getResource(namedGraphURI));
		}
	}


	@SuppressWarnings("unchecked")
	private void addSelectProperties(Query arq, Resource spinQuery) {
		if(arq.isDistinct()) {
			spinQuery.addProperty(SP.distinct, model.createTypedLiteral(true));
		}
		if(arq.isReduced()) {
			spinQuery.addProperty(SP.reduced, model.createTypedLiteral(true));
		}
		if(!arq.isQueryResultStar()) {
			List<Resource> members = new LinkedList<Resource>();
			VarExprList namedExprs = arq.getProject();
		    Iterator iter = namedExprs.getVars().iterator();
		    while(iter.hasNext()) {
		        Var var = (Var)iter.next();
		        Expr expr = namedExprs.getExpr(var) ;
		        if(expr == null) {
					String varName = (String) var.getName();
					Resource variable = getVariable(varName);
					members.add(variable);
		        }
		        else {
					String str = expr.toString();
					if(str.startsWith("count(")) {
						String varName = str.substring(7, str.length() - 1);
						Resource count = model.createResource(SP.Count);
						if(!str.contains("*")) {
							count.addProperty(SP.expression, getVariable(varName));
						}
						if(!var.isAllocVar()) {
							count.addProperty(SP.as, getVariable(var.getName()));
						}
						members.add(count);
					}
					else if(str.startsWith("sum(")) {
						String varName = str.substring(5, str.length() - 1);
						Resource count = model.createResource(SP.Sum);
						if(!str.contains("*")) {
							count.addProperty(SP.expression, getVariable(varName));
						}
						if(!var.isAllocVar()) {
							count.addProperty(SP.as, getVariable(var.getName()));
						}
						members.add(count);
					}
					else {
						throw new IllegalArgumentException("Expected count aggregation");
					}
		        }
		    }
			spinQuery.addProperty(SP.resultVariables, model.createList(members.iterator()));
		}
		addSolutionModifiers(arq, spinQuery);
	}
	
	
	@SuppressWarnings("unchecked")
	private void addSolutionModifiers(Query arq, Resource query) {
		long limit = arq.getLimit();
		if(limit != Query.NOLIMIT) {
			query.addProperty(SP.limit, query.getModel().createTypedLiteral(limit));
		}
		long offset = arq.getOffset();
		if(offset != Query.NOLIMIT) {
			query.addProperty(SP.offset, query.getModel().createTypedLiteral(offset));
		}
		
		List orderBy = arq.getOrderBy();
		if(orderBy != null && !orderBy.isEmpty()) {
			List<RDFNode> criteria = new LinkedList<RDFNode>();
			for(Object orderByRaw : orderBy) {
				SortCondition sortCondition = (SortCondition) orderByRaw;
				Expr expr = sortCondition.getExpression();
				RDFNode node = createExpression(expr);
				if(sortCondition.getDirection() == Query.ORDER_ASCENDING) {
					Resource asc = query.getModel().createResource(SP.Asc);
					asc.addProperty(SP.expression, node);
					criteria.add(asc);
				}
				else if(sortCondition.getDirection() == Query.ORDER_DESCENDING) {
					Resource desc = query.getModel().createResource(SP.Desc);
					desc.addProperty(SP.expression, node);
					criteria.add(desc);
				}
				else {
					criteria.add(node);
				}
			}
			query.addProperty(SP.orderBy, query.getModel().createList(criteria.iterator()));
		}
	}
	

	/**
	 * Creates a SPIN ElementList from a given ARQ Element pattern.
	 * @param pattern  the ARQ pattern to convert to SPIN
	 * @return a SPIN ElementList
	 */
	public ElementList createElementList(Element pattern) {
		final List<Resource> members = new LinkedList<Resource>();
		if(pattern != null) {
			pattern.visit(new AbstractElementVisitor() {
				
				private boolean first = true;
	
				@Override
				public void visit(ElementAssign assign) {
					RDFNode expression = createExpression(assign.getExpr());
					Variable variable = (Variable)getVariable(assign.getVar().getName()).as(Variable.class);
					members.add(SPINFactory.createLet(model, variable, expression));
				}
				
				
				@Override
				public void visit(ElementFilter filter) {
					RDFNode expression = createExpression(filter.getExpr());
					members.add(SPINFactory.createFilter(model, expression));
				}
	
				
				@Override
				public void visit(ElementGroup group) {
					if(first) {
						first = false;
						super.visit(group);
					}
					else {
						ElementList list = createElementList(group);
						members.add(list);
					}
				}


				@Override
				public void visit(ElementNamedGraph namedGraph) {
					Resource graphNameNode;
					Node nameNode = namedGraph.getGraphNameNode();
					if(nameNode.isVariable()) {
						graphNameNode = getVariable(nameNode.getName());
					}
					else {
						graphNameNode = model.getResource(nameNode.getURI());
					}
					Element element = namedGraph.getElement();
					RDFList elements = createElementList(element);
					NamedGraph ng = SPINFactory.createNamedGraph(model, graphNameNode, elements);
					members.add(ng);
				}


				@Override
				public void visit(ElementOptional optional) {
					Element element = optional.getOptionalElement();
					ElementList body = createElementList(element);
					Optional o = SPINFactory.createOptional(model, body);
					members.add(o);
				}
	
				
				public void visit(ElementPathBlock block) {
					visitElements(block.patternElts());
				}
	
				
				public void visit(ElementTriplesBlock el) {
					visitElements(el.patternElts());
				}


				@SuppressWarnings("unchecked")
				@Override
				public void visit(ElementUnion arqUnion) {
					List arqElements = arqUnion.getElements();
					List<RDFNode> elements = new LinkedList<RDFNode>();
					for(Object arqElement : arqElements) {
						RDFList element = createElementList((Element)arqElement);
						elements.add(element);
					}
					Union union = (Union) model.createResource(SP.Union).as(Union.class);
					union.addProperty(SP.elements, model.createList(elements.iterator()));
					members.add(union);
				}

				
				@SuppressWarnings("unchecked")
				private void visitElements(Iterator it) {
					while(it.hasNext()) {
						Object next = it.next();
						if(next instanceof TriplePath) {
							TriplePath path = (TriplePath)next;
							if(path.isTriple()) {
								next = path.asTriple();
							}
							else {
								Path p = path.getPath();
								Resource pathResource = createPath(p);
								Resource subject = (Resource) getNode(path.getSubject());
								RDFNode object = getNode(path.getObject());
								org.topbraid.spin.model.TriplePath triplePath = SPINFactory.createTriplePath(model, subject, pathResource, object);
								members.add(triplePath);
							}
						}
						if(next instanceof Triple) {
							Triple triple = (Triple) next;
							Resource subject = (Resource) getNode(triple.getSubject());
							Resource predicate = (Resource) getNode(triple.getPredicate());
							RDFNode object = getNode(triple.getObject());
							members.add(SPINFactory.createTriplePattern(model, subject, predicate, object));
						}
					}
				}
			});
		}
		return (ElementList) model.createList(members.iterator()).as(ElementList.class);
	}
	
	
	private RDFNode createExpression(Expr expr) {
		NodeValue constant = expr.getConstant();
		if(constant != null) {
			Node node = constant.asNode();
			return model.asRDFNode(node);
		}
		else {
			ExprVar var = expr.getNodeVar();
			if(var != null) {
				String varName = var.getVarName();
				return getVariable(varName);
			}
			else {
				ExprFunction function = expr.getFunction();
				Resource f = getFunction(function);
				FunctionCall call = SPINFactory.createFunctionCall(model, f);
				if(addPrefixes) {
					String ns = f.getNameSpace();
					if(ns != null && model.getNsURIPrefix(ns) == null) {
						Map<String,String> extras = ExtraPrefixes.getExtraPrefixes();
						for(String prefix : extras.keySet()) {
							if(ns.equals(extras.get(prefix))) {
								model.setNsPrefix(prefix, ns);
							}
						}
					}
				}
				List<RDFNode> params = createParameters(function);
				for(int i = 0; i < params.size(); i++) {
					RDFNode arg = params.get(i);
					Property predicate = model.getProperty(SP.NS + "arg" + (i + 1));
					call.addProperty(predicate, arg);
				}
				// call.addProperty(SP.arguments, model.createList(params.iterator()));
				return call;
			}
		}
	}
	
	
	private Resource createPath(Path path) {
		if(path instanceof P_Link) {
			P_Link link = (P_Link) path;
			Node node = link.getNode();
			return (Resource) model.asRDFNode(node);
		}
		else if(path instanceof P_Mod) {
			P_Mod mod = (P_Mod) path;
			Resource subR = createPath(mod.getSubPath());
			long max = mod.getMax();
			long min = mod.getMin();
			Resource r = model.createResource(SP.ModPath);
			r.addProperty(SP.subPath, subR);
			r.addProperty(SP.modMax, model.createTypedLiteral(max, XSD.integer.getURI()));
			r.addProperty(SP.modMin, model.createTypedLiteral(min, XSD.integer.getURI()));
			return r;
		}
		else if(path instanceof P_Alt) {
			P_Alt alt = (P_Alt) path;
			Resource path1 = createPath(alt.getLeft());
			Resource path2 = createPath(alt.getRight());
			Resource r = model.createResource(SP.AltPath);
			r.addProperty(SP.path1, path1);
			r.addProperty(SP.path2, path2);
			return r;
		}
		else if(path instanceof P_Reverse) {
			P_Reverse reverse = (P_Reverse) path;
			Resource r = model.createResource(SP.ReversePath);
			Resource path1 = createPath(reverse.getSubPath());
			r.addProperty(SP.subPath, path1);
			return r;
		}
		else if(path instanceof P_Seq) {
			P_Seq seq = (P_Seq) path;
			Resource path1 = createPath(seq.getLeft());
			Resource path2 = createPath(seq.getRight());
			Resource r = model.createResource(SP.SeqPath);
			r.addProperty(SP.path1, path1);
			r.addProperty(SP.path2, path2);
			return r;
		}
		else {
			throw new IllegalArgumentException("Unknown Path element: " + path);
		}
	}


	@SuppressWarnings("unchecked")
	private List<RDFNode> createParameters(ExprFunction function) {
		List<RDFNode> params = new LinkedList<RDFNode>();
		List args = function.getArgs();
		for(Object arg : args) {
			Expr argExpr = (Expr) arg;
			RDFNode param = createExpression(argExpr);
			params.add(param);
		}
		return params;
	}
	
	
	private Resource createHead(Template template) {
		final List<Resource> members = new LinkedList<Resource>();
		template.toString();
		template.visit(new AbstractTemplateVisitor() {
			@Override
			public void visit(TemplateTriple templateTriple) {
				Resource tripleTemplate = model.createResource(); // No SP.TripleTemplate needed
				Triple triple = templateTriple.getTriple();
				tripleTemplate.addProperty(SP.subject, getNode(triple.getSubject()));
				tripleTemplate.addProperty(SP.predicate, getNode(triple.getPredicate()));
				tripleTemplate.addProperty(SP.object, getNode(triple.getObject()));
				members.add(tripleTemplate);
			}
		});
		return model.createList(members.iterator());
	}
	
	
	/**
	 * Constructs a new SPIN Query from a given ARQ query, possibly
	 * with a URI.
	 * @param arq  the ARQ query
	 * @param uri  the URI of the new Query resource or null for a blank node
	 * @return the Query
	 */
	public org.topbraid.spin.model.Query createQuery(Query arq, String uri) {

		Resource spinQuery = model.createResource(uri);
		
		addNamedGraphClauses(arq, spinQuery);
		
		Resource where = createElementList(arq.getQueryPattern());
		spinQuery.addProperty(SP.where, where);
		
		if(arq.isAskType()) {
			spinQuery.addProperty(RDF.type, SP.Ask);
			return (Ask) spinQuery.as(Ask.class);
		}
		else if(arq.isConstructType()) {
			Resource head = createHead(arq.getConstructTemplate());
			spinQuery.addProperty(RDF.type, SP.Construct);
			spinQuery.addProperty(SP.templates, head);
			addSolutionModifiers(arq, spinQuery);
			return (Construct) spinQuery.as(Construct.class);
		}
		else if(arq.isSelectType()) {
			spinQuery.addProperty(RDF.type, SP.Select);
			Select select = (Select) spinQuery.as(Select.class);
			addSelectProperties(arq, spinQuery);
			addGroupBy(arq, spinQuery);
			return select;
		}
		else if(arq.isDescribeType()) {
			spinQuery.addProperty(RDF.type, SP.Describe);
			Describe describe = (Describe) spinQuery.as(Describe.class);
			addDescribeProperties(arq, spinQuery);
			addSolutionModifiers(arq, spinQuery);
			return describe;
		}
		throw new IllegalArgumentException("Unsupported SPARQL query type");
	}
	
	
	private Resource getFunction(ExprFunction function) {
		String symbol = function.getOpName();
		if(symbol == null) {
			symbol = function.getFunctionSymbol().getSymbol();
		}
		if(symbol != null) {
			StmtIterator it = symbolsModel.listStatements(null, SPIN.symbol, (RDFNode)null);
			while(it.hasNext()) {
				Statement s = it.nextStatement();
				if(s.getObject().isLiteral() && s.getString().toLowerCase().equals(symbol.toLowerCase())) {
					Resource f = s.getSubject();
					if(f.isURIResource()) {
						int count = 0;
						StmtIterator dit = f.listProperties(SPIN.constraint);
						while(dit.hasNext()) {
							dit.next();
							count++;
						}
						
						int argsCount = function.getArgs().size();
						if(argsCount == count) {
							it.close();
							return f;
						}
						else if(argsCount == 2 && SP.regex.equals(f)) {
							it.close();
							return f;
						}
					}
				}
			}
		}
		String iri = function.getFunctionIRI();
		if(iri != null) {
			return model.getResource(iri);
		}
		else {
			return null;
		}
	}
	
	
	private RDFNode getNode(Node node) {
		if(node.isVariable()) {
			String name = node.getName();
			return getVariable(name);
		}
		else {
			return model.asRDFNode(node);
		}
	}


	private Resource getVariable(String name) {
		Resource old = var2Resource.get(name);
		if(old != null) {
			return old;
		}
		else if(SPINPreferences.get().isCreateURIVariables()) {
			String uri = varNamespace + "_" + name;
			Resource var = model.createResource(uri, SP.Variable);
			var.addProperty(SP.varName, model.createTypedLiteral(name));
			var2Resource.put(name, var);
			return var;
		}
		else {
			Variable var = SPINFactory.createVariable(model, name); 
			var2Resource.put(name, var);
			return var;
		}
	}
	

	/**
	 * Gets the (optional) variable namespace.
	 * @return the variable namespace
	 */
	public String getVarNamespace() {
		return varNamespace;
	}
	
	
	/**
	 * Sets the variable namespace which is used to prevent the
	 * creation of too many blank nodes.
	 * @param value  the new namespace (might be null)
	 */
	public void setVarNamespace(String value) {
		this.varNamespace = value;
	}
}
