package org.topbraid.spin.inference;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.topbraid.spin.internal.JenaUtil;
import org.topbraid.spin.system.ARQFactory;
import org.topbraid.spin.system.SPINLabels;
import org.topbraid.spin.util.ProgressMonitor;
import org.topbraid.spin.util.QueryWrapper;
import org.topbraid.spin.util.SPINQueryFinder;
import org.topbraid.spin.util.SPINUtil;
import org.topbraid.spin.vocabulary.SPIN;

import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.query.QueryExecution;
import com.hp.hpl.jena.query.QuerySolutionMap;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.vocabulary.OWL;
import com.hp.hpl.jena.vocabulary.RDF;
import com.hp.hpl.jena.vocabulary.RDFS;


/**
 * A service to execute inference rules based on the spin:rule property.
 * 
 * @author Holger Knublauch
 */
public class SPINInferences {
	
	private static boolean isRootClass(Resource cls) {
		return RDFS.Resource.equals(cls) || OWL.Thing.equals(cls);
	}
	
	
	/**
	 * See the other run method for help - this is using spin:rule as rulePredicate.
	 * @param queryModel  the Model to query
	 * @param newTriples  the Model to add the new triples to 
	 * @param explanations  an optional object to write explanations to
	 * @param statistics  optional list to add statistics about which queries were slow
	 * @param singlePass  true to just do a single pass (don't iterate)
	 * @param monitor  an optional ProgressMonitor
	 * @return the number of iterations (1 with singlePass)
	 * @see #run(Model, Property, Model, SPINExplanations, List, boolean, ProgressMonitor)
	 */
	public static int run(
			Model queryModel, 
			Model newTriples,
			SPINExplanations explanations,
			List<SPINStatistics> statistics,
			boolean singlePass, 
			ProgressMonitor monitor) {
		return run(queryModel, SPIN.rule, newTriples, explanations, statistics, singlePass, monitor);
	}
	
	
	/**
	 * Iterates over all SPIN rules in a (query) Model and adds all constructed
	 * triples to a given Model (newTriples) until no further changes have been
	 * made within one iteration.
	 * Note that in order to iterate more than single pass, the newTriples Model
	 * must be a sub-model of the queryModel (which likely has to be an OntModel).
	 * The supplied rulePredicate is usually spin:rule, but can also be a sub-
	 * property of spin:rule to exercise finer control over which rules to fire.
	 * @param queryModel  the Model to query
	 * @param rulePredicate  the rule predicate (spin:rule or a sub-property thereof)
	 * @param newTriples  the Model to add the new triples to 
	 * @param explanations  an optional object to write explanations to
	 * @param statistics  optional list to add statistics about which queries were slow
	 * @param singlePass  true to just do a single pass (don't iterate)
	 * @param monitor  an optional ProgressMonitor
	 * @return the number of iterations (1 with singlePass)
	 */
	public static int run(
			Model queryModel,
			Property rulePredicate,
			Model newTriples,
			SPINExplanations explanations,
			List<SPINStatistics> statistics,
			boolean singlePass, 
			ProgressMonitor monitor) {
		Map<QueryWrapper, Map<String,RDFNode>> initialTemplateBindings = new HashMap<QueryWrapper, Map<String,RDFNode>>();
		Map<Resource,List<QueryWrapper>> cls2Query = SPINQueryFinder.getClass2QueryMap(queryModel, queryModel, rulePredicate, true, initialTemplateBindings, false);
		return run(queryModel, newTriples, cls2Query, initialTemplateBindings, explanations, statistics, singlePass, rulePredicate, monitor);
	}

	
	/**
	 * Iterates over a provided collection of SPIN rules and adds all constructed
	 * triples to a given Model (newTriples) until no further changes have been
	 * made within one iteration.
	 * Note that in order to iterate more than single pass, the newTriples Model
	 * must be a sub-model of the queryModel (which likely has to be an OntModel).
	 * @param queryModel  the Model to query
	 * @param newTriples  the Model to add the new triples to 
	 * @param class2Query  the map of queries to run (see SPINQueryFinder)
	 * @param templateBindings  initial template bindings (see SPINQueryFinder)
	 * @param explanations  an optional object to write explanations to
	 * @param statistics  optional list to add statistics about which queries were slow
	 * @param singlePass  true to just do a single pass (don't iterate)
	 * @param rulePredicate  the predicate used (e.g. spin:rule)
	 * @param monitor  an optional ProgressMonitor
	 * @return the number of iterations (1 with singlePass)
	 */
	public static int run(
			Model queryModel,
			Model newTriples,
			Map<Resource, List<QueryWrapper>> class2Query,
			Map<QueryWrapper, Map<String, RDFNode>> templateBindings,
			SPINExplanations explanations,
			List<SPINStatistics> statistics,
			boolean singlePass,
			Property rulePredicate,
			ProgressMonitor monitor) {
		if(rulePredicate.getModel() == null) {
			rulePredicate = queryModel.getProperty(rulePredicate.getURI());
		}
		int iteration = 1;
		boolean changed;
		do {
			Set<Statement> newRules = new HashSet<Statement>();
			changed = false;
			for(Resource cls : class2Query.keySet()) {
				List<QueryWrapper> arqs = class2Query.get(cls);
				for(QueryWrapper arqWrapper : arqs) {
					
					if(monitor != null) {
						StringBuffer sb = new StringBuffer("TopSPIN iteration ");
						sb.append(iteration);
						sb.append(" at ");
						sb.append(SPINLabels.get().getLabel(cls));
						monitor.subTask(sb.toString());
					}
	
					StringBuffer sb = new StringBuffer();
					sb.append("Inferred by ");
					sb.append(SPINLabels.get().getLabel(rulePredicate));
					sb.append(" at class ");
					sb.append(SPINLabels.get().getLabel(cls));
					sb.append(":\n\n" + arqWrapper.getText());
					String explanationText = sb.toString();
					Map<String,RDFNode> initialBindings = templateBindings.get(arqWrapper);
					Query arq = arqWrapper.getQuery();
					changed |= runQueryOnClass(arq, queryModel, newTriples, cls, true, initialBindings, statistics, explanations, explanationText, newRules, monitor);
					if(!isRootClass(cls)) {
						Set<Resource> subClasses = JenaUtil.getAllSubClasses(cls);
						for(Resource subClass : subClasses) {
							changed |= runQueryOnClass(arq, queryModel, newTriples, subClass, true, initialBindings, statistics, explanations, explanationText, newRules, monitor);
						}
					}
				}
			}
			iteration++;
			
			if(!newRules.isEmpty() && !singlePass) {
				for(Statement s : newRules) {
					SPINQueryFinder.add(class2Query, queryModel.asStatement(s.asTriple()), queryModel, true, templateBindings, false);
				}
			}
		}
		while(!singlePass && changed);
		
		return iteration - 1;
	}

	
	/**
	 * Runs a given Jena Query on a given instance and adds the inferred triples
	 * to a given Model.
	 * @param arq  the CONSTRUCT query to execute
	 * @param queryModel  the query Model
	 * @param newTriples  the Model to write the triples to
	 * @param instance  the instance to run the inferences on
	 * @param checkContains  true to only call add if a Triple wasn't there yet
	 * @param initialBindings  the initial bindings for arq or null
	 * @return true if changes were done (only meaningful if checkContains == true)
	 */
	public static boolean runQueryOnInstance(Query arq, Model queryModel, Model newTriples, Resource instance, boolean checkContains, Map<String,RDFNode> initialBindings) {
		boolean changed = false;
		QueryExecution qexec = ARQFactory.get().createQueryExecution(arq, queryModel);
		QuerySolutionMap bindings = new QuerySolutionMap();
		bindings.add(SPIN.THIS_VAR_NAME, instance);
		if(initialBindings != null) {
			for(String varName : initialBindings.keySet()) {
				RDFNode value = initialBindings.get(varName);
				bindings.add(varName, value);
			}
		}
		qexec.setInitialBinding(bindings);
		Model cm = qexec.execConstruct();
		StmtIterator cit = cm.listStatements();
		while(cit.hasNext()) {
			Statement s = cit.nextStatement();
			if(!checkContains || !queryModel.contains(s)) {
				changed = true;
				newTriples.add(s);
			}
		}
		return changed;
	}

	
	private static boolean runQueryOnClass(Query arq, Model queryModel, Model newTriples, Resource cls, boolean checkContains, Map<String,RDFNode> initialBindings, List<SPINStatistics> statistics, SPINExplanations explanations, String explanationText, Set<Statement> newRules, ProgressMonitor monitor) {
		
		// Check if query is needed at all
		if(isRootClass(cls) || queryModel.contains(null, RDF.type, cls)) {
			boolean changed = false;
			QueryExecution qexec = ARQFactory.get().createQueryExecution(arq, queryModel);
			QuerySolutionMap bindings = new QuerySolutionMap();
			if(!isRootClass(cls)) {
				bindings.add(SPINUtil.TYPE_CLASS_VAR_NAME, cls);
			}
			if(initialBindings != null) {
				for(String varName : initialBindings.keySet()) {
					RDFNode value = initialBindings.get(varName);
					bindings.add(varName, value);
				}
			}
			qexec.setInitialBinding(bindings);
			long startTime = System.currentTimeMillis();
			Model cm = qexec.execConstruct();
			Map<Resource,Resource> newInstances = new HashMap<Resource,Resource>();
			StmtIterator cit = cm.listStatements();
			while(cit.hasNext()) {
				Statement s = cit.nextStatement();
				if(!checkContains || !queryModel.contains(s)) {
					changed = true;
					newTriples.add(s);
					if(explanations != null) {
						explanations.put(s.asTriple(), explanationText);
					}
					
					// New rdf:type triple -> run constructors later
					if(RDF.type.equals(s.getPredicate()) && s.getObject().isResource()) {
						Resource subject = (Resource)s.getSubject().inModel(queryModel);
						newInstances.put(subject, s.getResource());
					}
					
					if(SPIN.rule.equals(s.getPredicate())) {
						newRules.add(s);
					}
				}
			}
			if(statistics != null) {
				long endTime = System.currentTimeMillis();
				long duration = (endTime - startTime);
				statistics.add(new SPINStatistics(arq, duration));
			}
			
			if(!newInstances.isEmpty()) {
				List<Resource> newRs = new ArrayList<Resource>(newInstances.keySet());
				SPINConstructors.construct(queryModel, newRs, newTriples, monitor);
			}
			
			return changed;
		}
		else {
			return false;
		}
	}
}
