package org.topbraid.spin.internal;

import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import com.hp.hpl.jena.enhanced.EnhGraph;
import com.hp.hpl.jena.graph.Graph;
import com.hp.hpl.jena.graph.compose.MultiUnion;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.NodeIterator;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.rdf.model.impl.PropertyImpl;
import com.hp.hpl.jena.rdf.model.impl.StmtIteratorImpl;
import com.hp.hpl.jena.vocabulary.RDF;
import com.hp.hpl.jena.vocabulary.RDFS;


/**
 * Some convenience methods to operate on Jena Models.
 * 
 * @author Holger Knublauch
 */
public class JenaUtil {


	private static void addTransitiveObjects(Set<Resource> resources, Resource subject, Property predicate) {
		addTransitiveObjects(resources, new HashSet<Resource>(), subject, predicate);
	}


	private static void addTransitiveObjects(Set<Resource> resources, Set<Resource> reached,
			Resource subject, Property predicate) {
		resources.add(subject);
		reached.add(subject);
		StmtIterator it = subject.listProperties(predicate);
		while (it.hasNext()) {
			RDFNode object = it.nextStatement().getObject();
			if (object instanceof Resource) {
				if (!reached.contains(object)) {
					addTransitiveObjects(resources, reached, (Resource)object, predicate);
				}
			}
		}
	}


	private static void addTransitiveSubjects(Set<Resource> reached, Resource object, Property predicate) {
		if (object != null) {
			reached.add(object);
			StmtIterator it = object.getModel().listStatements(null, predicate, object);
			while (it.hasNext()) {
				Resource subject = it.nextStatement().getSubject();
				if (!reached.contains(subject)) {
					addTransitiveSubjects(reached, (Resource)subject, predicate);
				}
			}
		}
	}

	
	/**
	 * Casts a Resource into a Property, without type checks etc.
	 * @param resource  the Resource to cast
	 * @return resource as an instance of Property
	 */
	public static Property asProperty(Resource resource) {
		if(resource instanceof Property) {
			return (Property) resource;
		}
		else {
			return new PropertyImpl(resource.asNode(), (EnhGraph)resource.getModel());
		}
	}


	/**
	 * Gets all instances of a given class and its subclasses.
	 * @param cls  the class to get the instances of
	 * @return the instances
	 */
	public static Set<Resource> getAllInstances(Resource cls) {
		Model model = cls.getModel();
		Set<Resource> classes = getAllSubClasses(cls);
		classes.add(cls);
		Set<Resource> results = new HashSet<Resource>();
		for(Resource subClass : classes) {
			StmtIterator it = model.listStatements(null, RDF.type, subClass);
			while (it.hasNext()) {
				results.add(it.nextStatement().getSubject());
			}
		}
		return results;
	}

	
	public static Set<Resource> getAllSubClasses(Resource cls) {
		return getAllTransitiveSubjects(cls, RDFS.subClassOf);
	}
	
	
	public static Set<Resource> getAllSubProperties(Property superProperty) {
		return getAllTransitiveSubjects(superProperty, RDFS.subPropertyOf);
	}

	
	public static Set<Resource> getAllSuperClasses(Resource cls) {
		return getAllTransitiveObjects(cls, RDFS.subClassOf);
	}
	
	
	public static Set<Resource> getAllSuperProperties(Property subProperty) {
		return getAllTransitiveObjects(subProperty, RDFS.subPropertyOf);
	}

	
	private static Set<Resource> getAllTransitiveObjects(Resource subject, Property predicate) {
		Set<Resource> set = new HashSet<Resource>();
		addTransitiveObjects(set, subject, predicate);
		set.remove(subject);
		return set;
	}

	
	private static Set<Resource> getAllTransitiveSubjects(Resource object, Property predicate) {
		Set<Resource> set = new HashSet<Resource>();
		addTransitiveSubjects(set, object, predicate);
		set.remove(object);
		return set;
	}

	
	public static Collection<Resource> getAllTypes(Resource instance) {
		Set<Resource> types = new HashSet<Resource>();
		StmtIterator it = instance.listProperties(RDF.type);
		while (it.hasNext()) {
			Resource type = it.nextStatement().getResource();
			types.add(type);
			types.addAll(getAllSuperClasses(type));
		}
		return types;
	}


	/**
	 * Overcomes a bug in Jena: if the base model does not declare a default namespace then the
	 * default namespace of an import is returned!
	 * 
	 * @param model the Model to operate on
	 * @param prefix the prefix to get the URI of
	 * @return the URI of prefix
	 */
	public static String getNsPrefixURI(Model model, String prefix) {
		if ("".equals(prefix) && model.getGraph() instanceof MultiUnion) {
			Graph baseGraph = ((MultiUnion)model.getGraph()).getBaseGraph();
			if(baseGraph != null) {
				return baseGraph.getPrefixMapping().getNsPrefixURI(prefix);
			}
			else {
				return model.getNsPrefixURI(prefix);
			}
		}
		else {
			return model.getNsPrefixURI(prefix);
		}
	}

	
	public static List<Statement> getStatementsList(StmtIterator it) {
		List<Statement> results = new LinkedList<Statement>();
		while(it.hasNext()) {
			results.add(it.nextStatement());
		}
		return results;
	}
	
	
	/**
	 * Gets a Set of all rdf:types of a given Resource. 
	 * @param instance  the instance Resource
	 * @return a Collection of type resources
	 */
	public static Collection<Resource> getTypes(Resource instance) {
		NodeIterator it = instance.getModel().listObjectsOfProperty(instance, RDF.type);
		Set<Resource> results = new HashSet<Resource>();
		while (it.hasNext()) {
			RDFNode node = it.nextNode();
			if (node instanceof Resource) {
				results.add((Resource)node);
			}
		}
		return results;
	}
	
	
	/**
	 * Checks whether a given Resource is an instance of a given type, or
	 * a subclass thereof.
	 * @param instance  the Resource to test
	 * @param type  the type
	 * @return true if resource has type type
	 */
	public static boolean hasIndirectType(Resource instance, Resource type) {
		if (instance.hasProperty(RDF.type, type)) {
			return true;
		}
		for(Resource subType : getAllSubClasses(type)) {
			if (instance.hasProperty(RDF.type, subType)) {
				return true;
			}
		}
		return false;
	}
	

	/**
	 * Checks whether a given property has a given (transitive) super property.
	 * @param subProperty  the sub-property
	 * @param superProperty  the super-property
	 * @return true if subProperty has superProperty (somewhere up the tree)
	 */
	public static boolean hasSuperProperty(Property subProperty, Property superProperty) {
		return getAllSuperProperties(subProperty).contains(superProperty);
	}
	
	
	/**
	 * Gets an Iterator over all Statements of a given property or its sub-properties
	 * at a given subject instance.  Note that the predicate and subject should be
	 * both attached to a Model to avoid NPEs.
	 * @param subject  the subject (may be null)
	 * @param predicate  the predicate
	 * @return a StmtIterator
	 */
	public static StmtIterator listAllProperties(Resource subject, Property predicate) {
		List<Statement> results = new LinkedList<Statement>();
		listAllProperties(subject, predicate, new HashSet<Property>(), results);
		return new StmtIteratorImpl(results.iterator());
	}
	
	
	private static void listAllProperties(Resource subject, Property predicate, Set<Property> reached,
			List<Statement> results) {
		reached.add(predicate);
		StmtIterator sit;
		Model model;
		if (subject != null) {
			sit = subject.listProperties(predicate);
			model = subject.getModel();
		}
		else {
			model = predicate.getModel();
			sit = model.listStatements(null, predicate, (RDFNode)null);
		}
		while (sit.hasNext()) {
			results.add(sit.nextStatement());
		}

		// Iterate into direct subproperties
		StmtIterator it = model.listStatements(null, RDFS.subPropertyOf, predicate);
		while (it.hasNext()) {
			Statement sps = it.nextStatement();
			if (!reached.contains(sps.getSubject())) {
				Property subProperty = asProperty(sps.getSubject());
				listAllProperties(subject, subProperty, reached, results);
			}
		}
	}
}
