package org.topbraid.spin.model.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.topbraid.spin.model.Element;
import org.topbraid.spin.model.ElementList;
import org.topbraid.spin.model.FunctionCall;
import org.topbraid.spin.model.SPINFactory;
import org.topbraid.spin.model.SolutionModifierQuery;
import org.topbraid.spin.model.print.PrintContext;
import org.topbraid.spin.system.ExtraPrefixes;
import org.topbraid.spin.util.SPINUtil;
import org.topbraid.spin.vocabulary.SP;

import com.hp.hpl.jena.enhanced.EnhGraph;
import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.rdf.model.Literal;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.vocabulary.RDF;
import com.hp.hpl.jena.vocabulary.RDFS;


public abstract class QueryImpl extends SPINResourceImpl implements SolutionModifierQuery {
	
	
	public QueryImpl(Node node, EnhGraph graph) {
		super(node, graph);
	}
	
	
	public String getComment() {
		return getString(RDFS.comment);
	}


	public List<String> getFrom() {
		return getStringList(SP.from);
	}


	public List<String> getFromNamed() {
		return getStringList(SP.fromNamed);
	}
	
	
	public Long getLimit() {
		return getLong(SP.limit);
	}


	public Long getOffset() {
		return getLong(SP.offset);
	}
	
	
	private String getPrefix(String namespace, PrintContext context) {
		String prefix = getModel().getNsURIPrefix(namespace);
		if(prefix == null && context.getUseExtraPrefixes()) {
			Map<String,String> extras = ExtraPrefixes.getExtraPrefixes();
			for(String extraPrefix : extras.keySet()) {
				String ns = extras.get(extraPrefix);
				if(namespace.equals(ns)) {
					return extraPrefix;
				}
			}
		}
		return prefix;
	}


	private List<String> getStringList(Property predicate) {
		List<String> results = new LinkedList<String>();
		StmtIterator it = listProperties(predicate);
		while(it.hasNext()) {
			RDFNode node = it.nextStatement().getObject();
			if(node.isLiteral()) {
				results.add(((Literal)node).getLexicalForm());
			}
			else if(node.isURIResource()) {
				results.add(((Resource)node).getURI());
			}
		}
		return results;
	}


	public ElementList getWhere() {
		Statement whereS = getProperty(SP.where);
		if(whereS != null) {
			Element element = SPINFactory.asElement(whereS.getResource());
			return (ElementList) element;
		}
		else {
			return null;
		}
	}


	public List<Element> getWhereElements() {
		return getElements(SP.where);
	}
	
	
	protected void printComment(PrintContext context) {
		String str = getComment();
		if(str != null) {
			String[] rows = str.split("\n");
			for(int i = 0; i < rows.length; i++) {
				context.print("# ");
				context.print(rows[i]);
				context.println();
			}
		}
	}
	
	
	protected void printPrefixes(PrintContext context) {
		if(context.getPrintPrefixes()) {
			Set<Resource> uriResources = SPINUtil.getURIResources(this);
			Set<String> namespaces = new HashSet<String>();
			for(Resource uriResource : uriResources) {
				String namespace = uriResource.getNameSpace();
				namespaces.add(namespace);
			}
			Map<String,String> prefix2Namespace = new HashMap<String,String>();
			for(String namespace : namespaces) {
				String prefix = getPrefix(namespace, context);
				if(prefix != null) {
					prefix2Namespace.put(prefix, namespace);
				}
			}
			List<String> prefixes = new ArrayList<String>(prefix2Namespace.keySet());
			Collections.sort(prefixes);
			for(String prefix : prefixes) {
				context.printKeyword("PREFIX");
				context.print(" ");
				context.print(prefix);
				context.print(": <");
				String ns = prefix2Namespace.get(prefix);
				context.print(ns);
				context.print(">");
				context.println();
			}
		}
	}


	protected void printStringFrom(PrintContext context) {
		for(String from : getFrom()) {
			context.println();
			context.printKeyword("FROM");
			context.print(" <");
			context.print(from);
			context.print(">");
		}
		for(String fromNamed : getFromNamed()) {
			context.println();
			context.printKeyword("FROM NAMED");
			context.print(" <");
			context.print(fromNamed);
			context.print(">");
		}
	}
	
	
	protected void printSolutionModifiers(PrintContext context) {
		List<RDFNode> orderBy = getList(SP.orderBy);
		if(!orderBy.isEmpty()) {
			context.println();
			context.printKeyword("ORDER BY");
			for(RDFNode node : orderBy) {
				if(node.isResource()) {
					Resource resource = (Resource) node;
					if(resource.hasProperty(RDF.type, SP.Asc)) {
						context.print(" ");
						context.printKeyword("ASC");
						context.print(" ");
						RDFNode expression = resource.getProperty(SP.expression).getObject();
						printOrderByExpression(context, expression);
					}
					else if(resource.hasProperty(RDF.type, SP.Desc)) {
						context.print(" ");
						context.printKeyword("DESC");
						context.print(" ");
						RDFNode expression = resource.getProperty(SP.expression).getObject();
						printOrderByExpression(context, expression);
					}
					else {
						context.print(" ");
						printOrderByExpression(context, node);
					}
				}
			}
		}
		Long limit = getLimit();
		if(limit != null) {
			context.println();
			context.printKeyword("LIMIT");
			context.print(" " + limit);
		}
		Long offset = getOffset();
		if(offset != null) {
			context.println();
			context.print("OFFSET");
			context.print(" " + offset);
		}
	}


	private void printOrderByExpression(PrintContext sb, RDFNode node) {
		
		if(node instanceof Resource) {
			Resource resource = (Resource) node;
			FunctionCall call = SPINFactory.asFunctionCall(resource);
			if(call != null) {
				sb.print("(");
				PrintContext pc = sb.clone();
				pc.setNested(true);
				call.print(pc);
				sb.print(")");
				return;
			}
		}
		
		printNestedExpressionString(sb, node, true);
	}


	protected void printWhere(PrintContext p) {
		p.printKeyword("WHERE");
		printNestedElementList(p, SP.where);
	}
}
