package org.topbraid.spin.model.impl;

import java.util.LinkedList;
import java.util.List;

import org.topbraid.spin.model.Element;
import org.topbraid.spin.model.ElementList;
import org.topbraid.spin.model.FunctionCall;
import org.topbraid.spin.model.SPINFactory;
import org.topbraid.spin.model.Variable;
import org.topbraid.spin.model.print.PrintContext;
import org.topbraid.spin.model.print.Printable;
import org.topbraid.spin.model.print.StringPrintContext;
import org.topbraid.spin.vocabulary.SP;

import com.hp.hpl.jena.enhanced.EnhGraph;
import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.rdf.model.Literal;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.RDFList;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.rdf.model.impl.ResourceImpl;
import com.hp.hpl.jena.sparql.util.FmtUtils;
import com.hp.hpl.jena.util.iterator.ExtendedIterator;
import com.hp.hpl.jena.vocabulary.RDF;


public abstract class SPINResourceImpl extends ResourceImpl implements Printable {
	
	/**
	 * One level of indentation (four spaces), used by toString methods
	 */
	public static final String INDENTATION = "    ";

	
	public SPINResourceImpl(Node node, EnhGraph graph) {
		super(node, graph);
	}
	
	
	public List<Element> getElements() {
		return getElements(SP.elements);
	}
	

	public List<Element> getElements(Property predicate) {
		List<Element> results = new LinkedList<Element>();
		for(RDFNode node : getList(predicate)) {
			if(node != null && node.isResource()) {
				results.add(SPINFactory.asElement((Resource)node));
			}
		}
		return results;
	}

	
	public List<RDFNode> getList(Property predicate) {
		Resource rawList = getResource(predicate);
		List<RDFNode> results = new LinkedList<RDFNode>();
		if(rawList != null) {
			RDFList list = (RDFList) rawList.as(RDFList.class);
			for(ExtendedIterator it = list.iterator(); it.hasNext(); ) {
				RDFNode node = (RDFNode) it.next();
				results.add(node);
			}
		}
		return results;
	}
	
	
	public Literal getLiteral(Property predicate) {
		Statement s = getProperty(predicate);
		if(s != null && s.getObject().isLiteral()) {
			return s.getLiteral();
		}
		else {
			return null;
		}
	}
	
	
	public Long getLong(Property predicate) {
		Literal literal = getLiteral(predicate);
		if(literal != null) {
			return literal.getLong();
		}
		else {
			return null;
		}
	}
	
	
	public RDFNode getRDFNode(Property predicate) {
		Statement s = getProperty(predicate);
		if(s != null) {
			return s.getObject();
		}
		else {
			return null;
		}
	}
	
	
	public Resource getResource(Property predicate) {
		Statement s = getProperty(predicate);
		if(s != null && s.getObject().isResource()) {
			return s.getResource();
		}
		else {
			return null;
		}
	}
	
	
	public List<Resource> getResources(Property predicate) {
		List<Resource> results = new LinkedList<Resource>();
		StmtIterator it = listProperties(predicate);
		while(it.hasNext()) {
			Statement s = it.nextStatement();
			if(s.getObject().isResource()) {
				results.add(s.getResource());
			}
		}
		return results;
	}
	
	
	public String getString(Property predicate) {
		Statement s = getProperty(predicate);
		if(s != null && s.getObject().isLiteral()) {
			return s.getString();
		}
		else {
			return null;
		}
	}
	
	
	public static boolean hasRDFType(Node node, EnhGraph graph, Resource type) {
		return graph.asGraph().contains(node, RDF.type.asNode(), type.asNode());
	}


	protected void printNestedElementList(PrintContext p) {
		printNestedElementList(p, SP.elements);
	}
	

	protected void printNestedElementList(PrintContext p, Property predicate) {
		p.print(" {");
		p.println();
		Resource elementsRaw = getResource(predicate);
		if(elementsRaw != null) {
			ElementList elements = (ElementList) elementsRaw.as(ElementList.class);
			p.setIndentation(p.getIndentation() + 1);
			elements.print(p);
			p.setIndentation(p.getIndentation() - 1);
		}
		p.printIndentation(p.getIndentation());
		p.print("}");
	}

	
	protected void printNestedExpressionString(PrintContext context, RDFNode node) {
		printNestedExpressionString(context, node, false);
	}
	
	
	protected void printNestedExpressionString(PrintContext p, RDFNode node, boolean force) {
		if(node instanceof Resource && SPINFactory.asVariable(node) == null) {
			Resource resource = (Resource) node;
			FunctionCall call = SPINFactory.asFunctionCall(resource);
			if(call != null) {
				PrintContext pc = p.clone();
				pc.setNested(true);
				call.print(pc);
				return;
			}
		}
		if(force) {
			p.print("(");
		}
		if(node instanceof Resource) {
			printVarOrResource(p, (Resource)node);
		}
		else {
			String str = FmtUtils.stringForNode(node.asNode(), getModel().getGraph().getPrefixMapping());
			p.print(str);
		}
		if(force) {
			p.print(")");
		}
	}

	
	public String toString() {
		StringPrintContext p = new StringPrintContext();
		print(p);
		return p.getString();
	}


	public static void printVarOrResource(PrintContext p, Resource resource) {
		Variable variable = SPINFactory.asVariable(resource);
		if(variable != null) {
			variable.print(p);
		}
		else if(resource.isURIResource()) {
			p.printURIResource(resource);
		}
		else {
			p.print("[]");
		}
	}
}
